from dataclasses import dataclass, field
from xml.dom import minidom
from copy import copy

from ..data_classes.tags import Tags

@dataclass
class osm_object_primitive():
    id: int | None = None
    visible: bool | None = None
    version: int | None = None
    changeset_id: int | None = None
    timestamp: str | None = None
    user_id: int | None = None
    tags: Tags = field(default_factory=Tags)

    def __str__(self):
        temp = f"{self.__class__.__name__}("
        for k in self.__dict__:
            temp += f"{k} = {getattr(self, k)}, "
        temp += ")"
        return temp

    def _to_xml(self, changeset_id, member_version=False, role="") -> minidom.Element:
        name = self.__class__.__name__.lower()
        root = minidom.Document()
        if member_version:
            element = root.createElement("member")
            element.setAttribute("type", name)
            element.setAttribute("ref", str(self.id))
            element.setAttribute("role", role)
            return element
        else:
            element = root.createElement(name)
            element.setAttribute("id",          str(self.id))
            element.setAttribute("version",     str(self.version))
            element.setAttribute("changeset",   str(changeset_id))
            return element
        
    def to_dict(self) -> dict[str, str]:
        """Returns a dictionary that corresponds to the attributes of the object. In addition, a 'type' key is added to specify the type of element.

        Returns:
            dict[str, str]: A dictionary that represents an object.
        """
        return_dict = copy(self.__dict__)
        return_dict.update({"type": self.__class__.__name__})
        return return_dict
    
    @classmethod
    def from_dict(cls, dict: dict[str, str]):
        """Creates an object from the data presented in the dictionary generated by the to_dict() method.

        Args:
            dict (dict[str, str]): Dictionary generated by the to_dict() method.

        Returns:
            Node | Way | Relation: The class object selected by the 'type' key in the dictionary.
        """
        if not dict.get("type"): 
            raise ValueError("No type key in the dictionary!")
        if dict["type"] != cls.__name__:
            raise ValueError(f'You used incorrect class to create object from given dictionary. Use {dict["type"]}.from_dict() instead of {cls.__name__}.from_dict()')
        node = cls()
        temp_dict = copy(dict)
        temp_dict.pop("type")
        node.__dict__ = temp_dict
        return node